import pandas as pd

from ..courtois.fmp_url import get_data_url

from datetime import datetime
from dateutil.relativedelta import relativedelta

def calculate_years(start_date: str):
    start_datetime = datetime.strptime(start_date, "%Y-%m-%d")
    now = datetime.now()
    difference = relativedelta(start_datetime, now)
    years = abs(difference.years)
    return years

def calculate_months(start_date: str):
    start_datetime = datetime.strptime(start_date, "%Y-%m-%d")
    now = datetime.now()
    difference = relativedelta(start_datetime, now)
    months = abs(difference.months) + abs(12*difference.years)
    return months

def get_fs_some(apikey: str='apikey', symbol: str='AAPL', period: str='annual', limit: int=10, finance: str='balance-sheet-statement'):
    """
    params:
        apikey -- to access data from FMP;
        symbol -- ticker of company;
        period -- either 'annual' or 'quarter';
        limit -- number of data points;
        finance -- can be 'balance-sheet-statement', or 'income-statement', or 'cash-flow-statement', or 'enterprise-values', 
            or 'ratios'
    return: list of dictionaries
    """
    url = (f'https://financialmodelingprep.com/api/v3/{finance}/{symbol}?period={period}&limit={limit}&apikey={apikey}')
    data = get_data_url(url=url)
    return data

def company_profile(apikey: str='apikey', symbol: str='AAPL'):
    """
    params:
        apikey -- to access data from FMP;
        symbol -- ticker of company;
    return: list of dictionaries
    """
    url = (f'https://financialmodelingprep.com/api/v3/profile/{symbol}?apikey={apikey}')
    data = get_data_url(url=url)
    return data

def get_fs(apikey: str='apikey', ticks: list=['AAPL', 'MSFT'], period: str = 'annual', 
           with_progress: bool = False, include_sic: bool = False, 
           bs_m=None, is_m=None, cf_m=None, ev_m=None, start_date: str= "N/A"):
    """
    params:
        apikey -- to access data from FMP;
        ticks -- list of tickers;
        period -- either 'annual' or 'quarter';
        with_progress -- if you want to visualize progress;
        include_sic -- if you want also extract sector and industry information
        bs_m -- the information from balance sheet like 'inventory', 'totalAssets', 'accountPayables' and so on...
                on default extracts maximum information from balance sheet
        is_m -- the information from balance sheet like 'revenue', 'grossProfit', 'eps', and so on...
                on default extracts maximum information from income statement
        cf_m -- the information from balance sheet like 'capitalExpenditure', 'dividendsPaid', 'operatingCashFlow' and so on...
                on default extracts maximum information from cash flow statement
        ev_m -- the information from balance sheet like 'enterpriseValue', 'marketCapitalization', 'numberOfShares', and 'stockPrice'
        start_date -- from which date we want to get financial statements
    return: data frame of financial statements
    """
    
    if start_date == "N/A":
        limit = 100;
    else:
        if period == "annual":
            limit = calculate_years(start_date=start_date) + 1;
        elif period == "quarter":
            limit = calculate_months(start_date=start_date) // 3 + 1;
        else:
            print("The period should be either 'annual' or 'quarter' ");
            return False;

    attributes = {
        'balance-sheet-statement': ['reportedCurrency', 'totalNonCurrentAssets', 'taxAssets', 'shortTermInvestments',
                         'longTermDebt', 'preferredStock', 'commonStock', 'totalLiabilitiesAndTotalEquity',
                         'deferredRevenue', 'goodwillAndIntangibleAssets', 'capitalLeaseObligations',
                         'totalStockholdersEquity', 'otherNonCurrentLiabilities', 'totalDebt', 'intangibleAssets',
                         'minorityInterest', 'otherAssets', 'propertyPlantEquipmentNet', 'cashAndShortTermInvestments',
                         'symbol', 'totalEquity', 'totalLiabilities', 'otherNonCurrentAssets', 'otherCurrentLiabilities',
                         'retainedEarnings', 'totalNonCurrentLiabilities', 'othertotalStockholdersEquity',
                         'deferredRevenueNonCurrent', 'accountPayables', 'totalCurrentLiabilities', 'totalCurrentAssets',
                         'longTermInvestments', 'deferredTaxLiabilitiesNonCurrent', 'netReceivables', 'taxPayables',
                         'shortTermDebt', 'cashAndCashEquivalents', 'inventory', 'otherCurrentAssets', 'netDebt', 'date',
                         'otherLiabilities', 'goodwill', 'totalInvestments', 'totalLiabilitiesAndStockholdersEquity',
                         'totalAssets', 'accumulatedOtherComprehensiveIncomeLoss'],
        'income-statement': ['incomeBeforeTaxRatio', 'incomeTaxExpense', 'totalOtherIncomeExpensesNet', 'revenue', 
                         'costOfRevenue', 'otherExpenses', 'netIncome', 'generalAndAdministrativeExpenses',
                         'ebitdaratio', 'eps', 'symbol', 'operatingIncomeRatio', 'sellingGeneralAndAdministrativeExpenses',
                         'weightedAverageShsOut', 'operatingExpenses', 'weightedAverageShsOutDil', 'interestExpense',
                         'incomeBeforeTax', 'epsdiluted', 'depreciationAndAmortization', 'ebitda',
                         'sellingAndMarketingExpenses', 'date', 'grossProfit', 'costAndExpenses', 'grossProfitRatio',
                         'researchAndDevelopmentExpenses', 'netIncomeRatio', 'interestIncome', 'operatingIncome'],
        'cash-flow-statement': ['otherWorkingCapital', 'commonStockRepurchased', 'investmentsInPropertyPlantAndEquipment',
                         'dividendsPaid', 'otherNonCashItems', 'netCashUsedForInvestingActivites', 'freeCashFlow',
                         'cashAtEndOfPeriod', 'commonStockIssued', 'effectOfForexChangesOnCash', 'symbol',
                         'otherFinancingActivites', 'netCashProvidedByOperatingActivities',
                         'netCashUsedProvidedByFinancingActivities', 'deferredIncomeTax', 'stockBasedCompensation',
                         'changeInWorkingCapital', 'netChangeInCash', 'debtRepayment', 'acquisitionsNet', 'accountsPayables',
                         'operatingCashFlow', 'date', 'accountsReceivables', 'salesMaturitiesOfInvestments',
                         'cashAtBeginningOfPeriod', 'otherInvestingActivites', 'purchasesOfInvestments', 'capitalExpenditure'],
        'enterprise-values': ['enterpriseValue', 'marketCapitalization', 'numberOfShares', 'stockPrice', 'date', 'symbol']
    }
    
    metrics = {'balance-sheet-statement': bs_m, 'income-statement': is_m,
               'cash-flow-statement': cf_m, 'enterprise-values': ev_m}
    fss = ['balance-sheet-statement', 'income-statement', 'cash-flow-statement', 'enterprise-values']
    attributes_fss = {}

    for fs in fss:
        if (metrics[fs] == None):
            attributes_fs = attributes[fs]
        elif (metrics[fs] == []):
            attributes_fs = []
        else:
            attributes_fs = set(metrics[fs]);
            attributes_fs.add('date');
            attributes_fs.add('symbol');
            attributes_fs = list(attributes_fs);
    
        attributes_fss[fs] = attributes_fs

    data_frames = []

    if with_progress:
        i = 0;
        n = len(ticks);

    fss = ['balance-sheet-statement', 'income-statement', 'cash-flow-statement', 'enterprise-values']

    for ticker in ticks:

        inventory_frames = []

        if with_progress:
            i += 1;
        
        for fs in fss:

            if (attributes_fss[fs] == []):
                continue;
            else:
                ticker_data = get_fs_some(apikey=apikey, symbol=ticker, period=period, limit=limit, finance=fs)

                try:
                    data_df = pd.DataFrame(ticker_data)
                    data_df = data_df[list(set(attributes_fss[fs]).intersection(set(data_df.columns)))]
                    data_df.drop_duplicates(subset=['symbol', 'date'], inplace=True)
                    data_df.set_index(['symbol', 'date'], inplace=True)
                    inventory_frames.append(data_df)
                except:
                    pass

        if (inventory_frames == []):
            continue;
        else:
            try:
                data_inventory_stock_q = pd.concat(inventory_frames, axis=1)
                data_frames.append(data_inventory_stock_q)

                if with_progress:
                    print(f"With ticker (fs) -- {ticker} : {str(i)} / {str(n)} with limit -- {str(limit)} ")
            except:
                continue
    
    # print(data_frames)

    if start_date == "N/A":
        df = pd.concat(data_frames)
    else:
        df = pd.concat(data_frames)
        df.reset_index(inplace=True)

        df = df[df['date'] >= start_date]

        df.set_index(['symbol', 'date'], inplace=True)

    if include_sic:
        tickers_downloaded = list(set(df.index.get_level_values(0)))

        df.reset_index(inplace=True)

        df['sector'] = 'other';
        df['industry'] = 'other';

        if with_progress:
            i = 0;
            n = len(tickers_downloaded);

        for ticker in tickers_downloaded:
            if with_progress:
                i += 1;

            company_profile = company_profile(apikey=apikey, symbol=ticker);

            if company_profile != []:
                df.loc[df['symbol'] == ticker, 'sector'] = company_profile[0].get('sector', 'other');
                df.loc[df['symbol'] == ticker, 'industry'] = company_profile[0].get('industry', 'other');

                if with_progress:
                    print(f"For ticker (sic) -- {ticker} : {str(i)} / {str(n)}");

        df.set_index(['symbol', 'date'], inplace=True)

    return df


import pandas as pd

# import fmpsdk
from ..courtois.fmp_url import get_data_url
from get_collection_fs import calculate_years, calculate_months, get_fs_some

from datetime import datetime
from dateutil.relativedelta import relativedelta

# def get_ratios_fmp(apikey: str, symbol: str, period: str='annual', limit: int=10):
#     url = (f'https://financialmodelingprep.com/api/v3/ratios/{symbol}?period={period}&limit={limit}&apikey={apikey}')
#     data = get_data_url(url=url)
#     return data

def get_ratios(apikey: str='apikey', ticks: list=['AAPL', 'MSFT'], period: str = 'annual', 
           with_progress: bool = False, ratios=[], start_date: str= "N/A"):
    """
    params:
        apikey -- to access data from FMP;
        ticks -- list of tickers;
        period -- either 'annual' or 'quarter';
        with_progress -- if you want to visualize progress;
        ratios -- list of ratios;
        start_date -- from which date we want to get financial statements
    return: data frame of ratios
    """
    
    if start_date == "N/A":
        limit = 100;
    else:
        if period == "annual":
            limit = calculate_years(start_date=start_date) + 1;
        elif period == "quarter":
            limit = calculate_months(start_date=start_date) // 3 + 1;
        else:
            print("The period should be either 'annual' or 'quarter' ");
            return False;

    if (ratios == []):
        attributes_ratios = ['symbol', 'date', 
                             'daysOfSalesOutstanding', 'daysOfInventoryOutstanding', 'daysOfPayablesOutstanding',
                             'cashConversionCycle', 'cashRatio', 
                             'grossProfitMargin', 'inventoryTurnover', 'freeCashFlowPerShare']
    else:
        attributes_ratios = set(ratios);
        attributes_ratios.add('date');
        attributes_ratios.add('symbol');
        attributes_ratios = list(attributes_ratios);

    data_frames = []

    if with_progress:
        i = 0;
        n = len(ticks);

    for ticker in ticks:

        if with_progress:
            i += 1;

        ticker_ratios = get_fs_some(apikey=apikey, symbol=ticker, period=period, limit=limit, finance='ratios')

        if ((ticker_ratios == [])):
            continue

        if ((type(ticker_ratios) == dict) and (ticker_ratios.get('status', '') == 404)):
            continue

        data_ratios = pd.DataFrame(ticker_ratios)
        

        data_ratios = data_ratios[list(set(attributes_ratios).intersection(set(list(data_ratios.columns))))]
        if (('symbol' not in list(data_ratios.columns)) or ('date' not in list(data_ratios.columns))):
            continue
        data_ratios.drop_duplicates(subset=['symbol', 'date'], inplace=True)
        data_ratios.set_index(['symbol', 'date'], inplace=True)

        try:

            data_frames.append(data_ratios)

            if with_progress:
                print(f"With ticker (fs) -- {ticker} : {str(i)} / {str(n)} with limit -- {str(limit)} ")

        except:
            continue;
    
    # print(data_frames)

    if start_date == "N/A":
        df = pd.concat(data_frames)
    else:
        df = pd.concat(data_frames)
        df.reset_index(inplace=True)

        df = df[df['date'] >= start_date]

        df.set_index(['symbol', 'date'], inplace=True)

    return df


import pandas as pd

import plotly.graph_objects as go

def format_string(input_string):
    formatted_string = ''
    for char in input_string:
        if char.isupper():
            formatted_string += ' '
        formatted_string += char
    formatted_string = formatted_string.replace('_', ' ')
    formatted_string = formatted_string.strip()
    formatted_string = formatted_string.title()
    return formatted_string

def line_graph(field, data):
    x_values = data['date']
    y_values = data[field]
    
    trace = go.Scatter(
    x=x_values,
    y=y_values,
    mode='lines',
    name='Line Graph'
    )
    
    layout = go.Layout(
    title='Line Graph Example',
    xaxis=dict(title='date'),
    yaxis=dict(title=format_string(field))
    )
    
    return go.Figure(data=[trace], layout=layout)




import plotly.express as px

from plotly.subplots import make_subplots
import plotly.graph_objects as go

def plot_with_margin(data_frame, x: str, y: str, color: str, palette: dict, error: str=None, upper_bound: str=None, lower_bound: str=None):

    if (~bool(error)):
        if ((~bool(lower_bound)) or (~bool(upper_bound))):
            print("There no error parameter you can use standard px.line")
            return False
    else:
        lower_bound='lower_bound'
        data_frame[lower_bound] = data_frame[y] - data_frame[error]/2
        upper_bound='upper_bound'
        data_frame[upper_bound] = data_frame[y] + data_frame[error]/2

    colors = set(data_frame[color])

    if (colors != set(palette.keys())):
        print(f"The {color} column and the keys of palette are not same")
        return False

    go_scatters = []

    for colour in colors:
        if colour[:3]=='rgb':
            fillcolour = 'rgba' + colour[3:-1] + ', 0.3)'

        go_scatters.append(
            go.Scatter(
                name=str(colour), x=data_frame[x], y=data_frame[y], mode='lines',
                line=dict(color=palette[colour])
            ))
        go_scatters.append(
            go.Scatter(
                name='Upper Bound', x=data_frame[x], y=data_frame[upper_bound], mode='lines',
                marker=dict(color="#444"), line=dict(width=0), showlegend=False
            ))
        go_scatters.append(
            go.Scatter(
                name='Lower Bound', x=data_frame[x], y=data_frame[lower_bound], mode='lines',
                marker=dict(color="#444"), line=dict(width=0), fillcolor=fillcolour,
                fill='tonexty', showlegend=False
            ))

    fig = go.Figure(go_scatters)
    fig.update_layout(
        yaxis_title=y,
        title=f"Continuous, {y}")
    
    return fig



import pandas as pd

from ..courtois.fmp_url import get_data_url
from .get_collection_fs import calculate_years

from datetime import datetime
from dateutil.relativedelta import relativedelta

def get_stocks_fmp(apikey: str='apikey', symbols: list=['AAPL', 'MSFT', 'GOOG', '018260.KS', 'SONY'], start_date: str='2023-10-01'):
    """
    params:
        apikey -- to access data from FMP;
        symbols -- list of tickers of companies, should be not bigger than 5;
        start_date -- the date from which to collect data, not early than 1 year before
    return: list of dictionaries
    """
    symbols_txt = ",".join(symbols)
    url = (f'https://financialmodelingprep.com/api/v3/historical-price-full/{symbols_txt}?from={start_date}&apikey={apikey}')
    data = get_data_url(url=url)
    return data

def get_stocks(apikey: str='apikey', symbols: list=['GOOG', 'MSFT', 'AMZN'], start_date: str='1986-01-01', 
                step:int=5, with_progress: bool=False):
    """
    params:
        apikey -- to access data from FMP;
        symbols -- list of tickers of companies, can be bigger than 5;
        start_date -- the date from which to collect data, if start_date is early than 1 year before then step should be 1;
        with_progress -- if you want to monitor the progress;
        step -- can either 5 or 4 or 3 or 2 or 1
    return: list of dictionaries
    """
    data_frames = []

    symbols1 = symbols.copy()

    if with_progress:
        n = len(symbols) // step
        i = 0

    while(len(symbols1) >= 1):

        if calculate_years(start_date) > 1:
            step = 1

        if (len(symbols1) > step):
            symbols0 = symbols1[:step]
            symbols1 = symbols1[step:]
        else:
            symbols0 = symbols1
            symbols1 = []

        data = get_stocks_fmp(apikey=apikey, symbols=symbols0, start_date=start_date)
            
        data_frames0 = []

        if step > 1:
            for d in data['historicalStockList']:
                df = pd.DataFrame(d['historical'])
                df['symbol'] = d['symbol']
                data_frames0.append(df)
        else:
            df = pd.DataFrame(data['historical'])
            df['symbol'] = data['symbol']
            data_frames0.append(df)


        df = pd.concat(data_frames0)

        data_frames.append(df)

        if with_progress:
            i += 1
            print(i, '/' ,n)

    df = pd.concat(data_frames)

    return df


import pandas as pd

# import fmpsdk

# try:
#     # For Python 3.0 and later
#     from urllib.request import urlopen
# except ImportError:
#     # Fall back to Python 2's urllib2
#     from urllib2 import urlopen

import plotly.graph_objects as go

from ..sapar.get_collection_fs import get_fs_some
from ..sapar.get_collection_ratios import get_ratios_fmp
from fmp_url import get_data_url

def data_merge(apikey, ticker, period): # get and merge data from BSS; CFS; IS; and ratios - input period and ticker
    
    symbol: str = ticker
        
    data_BSS = pd.DataFrame(get_fs_some(apikey=apikey, symbol=symbol, period=period, finance='balance-sheet-statement'))
    data_CFS = pd.DataFrame(get_fs_some(apikey=apikey, symbol=symbol, period=period, finance='income-statement'))
    data_IS = pd.DataFrame(get_fs_some(apikey=apikey, symbol=symbol, period=period, finance='cash-flow-statement'))
    data_ratios = pd.DataFrame(get_ratios_fmp(apikey=apikey, symbol=symbol, period=period))

    data_CFS.pop('inventory')
    data_CFS.pop('netIncome')
    data_CFS.pop('depreciationAndAmortization')

    data_annual = data_BSS.copy()
    data_annual = data_annual.merge(data_CFS, how='inner', on=['date', 'cik', 'period',
                                                               'reportedCurrency', 'symbol',
                                                               'link', 'finalLink', 'fillingDate',
                                                               'acceptedDate', 'calendarYear'])
    data_annual = data_annual.merge(data_IS, how='inner', on=['date', 'cik', 'period',
                                                              'reportedCurrency', 'symbol', 'link',
                                                              'finalLink', 'fillingDate',
                                                              'acceptedDate', 'calendarYear'])
    data_annual = data_annual.merge(data_ratios, how='inner', on=['date', 'period', 'symbol'])

    return data_annual

# def get_data_url(url): #Get data using API link - NOT fmpsdk lib
#     response = urlopen(url, cafile=certifi.where())
#     data = response.read().decode("utf-8")
#     data = json.loads(data)
#     return data

def get_competitors(apikey, ticker): #Get competitors - NOT fmpsdk lib
    
    url = (f'https://financialmodelingprep.com/api/v4/stock_peers?symbol={ticker}&apikey={apikey}')
    
    data = get_data_url(url)
    return data[0]['peersList']

def average_calc(apikey, ticker, field, period): #input a value and return avg competitors - NOT fmpsdk lib
    competitors = get_competitors(ticker)
    
    df_rtr = pd.DataFrame()
    for comp in competitors:
        try:
            data_comp = data_merge(apikey, comp, period)
            data_comp = data_comp[['calendarYear', 'period', field]]
            df_rtr = pd.concat([df_rtr, data_comp], ignore_index=True)
        except:
            pass
        
    averaged_df = df_rtr.groupby('calendarYear')[field].mean().reset_index()

    return averaged_df

def format_string(input_string):
    formatted_string = ''
    for char in input_string:
        if char.isupper():
            formatted_string += ' '
        formatted_string += char
    formatted_string = formatted_string.replace('_', ' ')
    formatted_string = formatted_string.strip()
    formatted_string = formatted_string.title()
    return formatted_string

def line_graph(field,data):
    x_values=data['date']
    y_values=data[field]
    
    trace = go.Scatter(
    x=x_values,
    y=y_values,
    mode='lines',
    name='Line Graph'
    )
    
    layout = go.Layout(
    title='Line Graph Example',
    xaxis=dict(title='date'),
    yaxis=dict(title=format_string(field))
    )
    
    return go.Figure(data=[trace], layout=layout)


import pandas as pd
import numpy as np

import statsmodels.formula.api as smf

def slope_corr_ols(metric0: float='inventory'):
    def slope_corr(data_frame):
        data_frame = data_frame.dropna()

        if len(data_frame) <= 0:
            return pd.Series([None, None], index=["slope", "correlation"])
    
        cols = list(set(data_frame.columns).difference([metric0]))
        col = cols[0]

        inventory_model = smf.ols(formula=col + ' ~ ' + metric0, data=data_frame)
        inventory_model = inventory_model.fit()
    
        return pd.Series([inventory_model.params[metric0], 
                      data_frame[metric0].corr(data_frame[col])], 
                      index=["slope", "correlation"])

    return slope_corr